from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Union
from uuid import uuid4

from numpy import float32, ndarray
from sentence_transformers import SentenceTransformer


@dataclass
class EmbedModel:
    name: str
    model: SentenceTransformer = field(init=False)
    tokenizer: Callable = field(init=False)

    def __post_init__(self) -> None:
        self.model = SentenceTransformer(self.name)
        self.tokenizer = self.model[0].tokenizer

    def tokenize(self, input: str) -> List[str]:
        return self.tokenizer.tokenize(input)

    def encode(self, input: str) -> List[int]:
        return self.model.encode(input)


@dataclass
class Document:
    text: str
    metadata: Optional[Dict[str, Union[str, int, List, Dict]]] = None
    id: Union[str, int] = field(default_factory=lambda: uuid4().hex)


@dataclass
class Collection:
    name: str
    documents: Optional[List[Document]] = None
    embeddings: Optional[List[ndarray[float32]]] = None

    def add(
        self, documents: List[Document], embeddings: List[ndarray[float32]]
    ) -> None:
        self.documents.extend(documents)
        self.embeddings.extend(embeddings)

    def update(
        self,
        index: Union[str, int],
        text: str,
        embedding: List[ndarray[float32]],
        metadata: Dict[str, Union[str, int, List, Dict]],
    ) -> None:
        self.documents[index].text = text
        self.documents[index].metadata = metadata
        self.embeddings[index] = embedding

    def remove(self, index: Union[str, int]) -> None:
        del self.documents[index]
        del self.embeddings[index]

    def get_indices(
        self, target_ids: Union[List[Union[str, int]], str, int]
    ) -> Union[List[int], int]:
        if not isinstance(target_ids, (list, str, int)):
            raise ValueError("Invalid target id. Ensure target documents exists.")
        try:
            if isinstance(target_ids, list):
                return [
                    idx
                    for idx, doc in enumerate(self.documents)
                    if doc.id in target_ids
                ]
            else:
                return [
                    idx
                    for idx, doc in enumerate(self.documents)
                    if doc.id == target_ids
                ][0]
        except IndexError:
            raise IndexError("One or multiple target ids within  do not exist.")



import numpy as np
import pytest

from iotadb.iotadb import (OPSET_LOOKUP, Collection, Document, EmbedModel,
                           IotaDB)


@pytest.mark.parametrize(
    "metric, embed_model, persist, persist_dir",
    [
        ("cosine", "all-mpnet-base-v2", False, None),
        ("cosine", "all-mpnet-base-v2", True, "data/"),
        ("dot", "all-mpnet-base-v2", False, None),
        ("dot", "all-mpnet-base-v2", True, "data/"),
        ("euclidean", "all-mpnet-base-v2", False, None),
        ("euclidean", "all-mpnet-base-v2", True, "data/"),
    ],
)
def test_database_init(metric, embed_model, persist, persist_dir):
    db = IotaDB(
        metric=metric, embed_model=embed_model, persist=persist, persist_dir=persist_dir
    )

    assert db.sim_func == OPSET_LOOKUP[metric]
    assert isinstance(db.embed_model, EmbedModel)
    assert db.persist == persist
    assert isinstance(db.persist, bool)
    assert db.persist_dir == persist_dir
    assert isinstance(db.persist_dir, str) or db.persist_dir is None


def test_create_collection(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    assert isinstance(default_database.get_collection(), Collection)


# def test_load_collection():
#     pass


def test_get_document(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    doc1 = default_database.get_documents(ids=[0])
    doc2 = default_database.get_documents(ids=[0], include_embeddings=True)

    assert isinstance(doc1, list)
    assert isinstance(doc1[0], Document)
    assert isinstance(doc2, list)
    assert isinstance(doc2[0], tuple)
    assert isinstance(doc2[0][0], Document)
    assert isinstance(doc2[0][1], np.ndarray)
    assert doc2[0][1].dtype == np.float32


def test_add_documents(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    sample_doc_len = len(sample_documents)

    new_docs = [
        Document(id=5, text="The quick brown fox jumps over the lazy dog"),
        Document(id=6, text="The five boxing wizards jump quickly."),
    ]

    default_database.add_documents(documents=new_docs)
    collection = default_database.get_collection()

    assert len(collection.documents) == sample_doc_len + len(new_docs)


def test_update_document(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    default_database.update_document(id=0, new_text="hello world")
    collection = default_database.get_collection()
    assert collection.documents[0].text == "hello world"


def test_remove_document(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    sample_doc_len = len(sample_documents)
    default_database.remove_document(id=0)
    collection = default_database.get_collection()

    assert len(collection.documents) == sample_doc_len - 1


def test_search(default_database, sample_documents):
    default_database.create_collection(
        name="test_collection", documents=sample_documents
    )

    results1 = default_database.search(query="a dancing cat")
    results2 = default_database.search(query="a dancing cat", return_similarities=True)

    assert isinstance(results1, list)
    assert isinstance(results1[0], Document)
    assert isinstance(results2, list)
    assert isinstance(results2[0][0], Document)
    assert results2[0][1].dtype == np.float32





import os
from pickle import HIGHEST_PROTOCOL, dump, load
from typing import Dict, Iterable, List, Literal, Optional, Tuple, Union

from numpy import argsort, float32, vstack

from .metrics import OPSET_LOOKUP
from .schemas import Collection, Document, EmbedModel


class IotaDB:
    def __init__(
        self,
        metric: Literal["dot", "cosine", "euclidean"] = "cosine",
        embed_model: str = "all-mpnet-base-v2",
        persist: bool = False,
        persist_dir: Optional[str] = None,
    ) -> None:
        """
        Initializes database instance

        Args:
            metric: distance metric used for vector search
            embed_mode: name of sentence-transformer model
            persist: whether to persis data to disk
            persis_dir: directory to save to
        """

        if metric not in OPSET_LOOKUP.keys():
            raise NotImplementedError("Algorithm not implemented.")
        if persist and persist_dir is None:
            raise ValueError("Path must be specified when persisting.")

        self.sim_func = OPSET_LOOKUP[metric]
        self.embed_model = EmbedModel(name=embed_model)
        self.tokenizer = self.embed_model.tokenizer
        self.persist = persist
        self.persist_dir = persist_dir
        self._collection = None

        if self.persist:
            os.makedirs(self.persist_dir, exist_ok=True)

    def create_collection(
        self,
        name: str,
        documents: Optional[List[Document]] = None,
    ) -> None:
        """
        Creates a collection with or without documents,

        Args:
            name: name of collection
            documents: list of documents to create collection with
        """
        documents = [] if documents is None else documents
        embeddings = (
            []
            if documents is None
            else [self._get_embedding(doc.text) for doc in documents]
        )

        self._collection = Collection(
            name=name, documents=documents, embeddings=embeddings
        )

        if self.persist:
            self._save_to_disk(f"{self._collection.name}.pickle")

    def load_collection(self, file_path: str) -> None:
        """
        Loads a collection from file

        Args:
            file_path: path to collection file
        """
        with open(file_path, "rb") as f:
            self._collection = load(f)

    def get_collection(self) -> Collection:
        return self._collection 

    def add_documents(self, documents: List[Document]) -> None:
        """
        Adds one or more documents to a collection

        Args:
            documents: a list of documents to add to an existing collection
        """
        if self._collection is None:
            raise Exception(
                "No collection exists. Create a collection before adding documents."
            )

        embeddings = [self._get_embedding(doc.text) for doc in documents]
        self._collection.add(documents=documents, embeddings=embeddings)

        if self.persist:
            self._save_to_disk(f"{self._collection.name}.pickle")

    def get_documents(
        self, ids: List[Union[str, int]], include_embeddings: bool = False
    ) -> Union[List[Document], Iterable[Tuple]]:
        '''
        Gets one or more documents from a collection

        Args:
            ids: list of document ids to retrieve from an existing collection
            include_embeddings: whether to include embeddings during retrieval
        '''
        if self._collection is None:
            raise Exception("No existing collection. Create one first.")

        indices = self._collection.get_indices(target_ids=ids)
        documents = [self._collection.documents[i] for i in indices]

        if include_embeddings:
            embeddings = [self._collection.embeddings[i] for i in indices]
            return list(zip(documents, embeddings))

        return documents

    def update_document(
        self,
        id: Union[str, int],
        new_text: str,
        new_metadata: Optional[Dict[str, Union[str, int, List, Dict]]] = None,
    ) -> None:
        '''
        Updates contents of a document given an id

        Args:
            id: a document's id
            new_text: new text to update a document with
            new_metadata: new metadata to update a document with
        '''
        if self._collection is None:
            raise Exception("No existing collection. Create one first.")

        index = self._collection.get_indices(target_ids=id)
        new_embedding = self._get_embedding(new_text)

        self._collection.update(
            index=index, text=new_text, embedding=new_embedding, metadata=new_metadata
        )

        if self.persist:
            self._save_to_disk(f"{self._collection.name}.pickle")

    def remove_document(self, id: Union[str, int]) -> None:
        '''
        Removes a document given an id 

        Args:
            id: a document's id
        '''
        if self._collection is None:
            raise Exception("No existing collection. Create one first.")

        index = self._collection.get_indices(target_ids=id)
        self._collection.remove(index=index)

        if self.persist:
            self._save_to_disk(f"{self._collection.name}.pickle")

    def search(
        self,
        query: str,
        top_k: int = 10,
        return_similarities: bool = False,
    ) -> List[Union[Document, Tuple[Document, float32]]]:
        """
        Performs a naive brute-force search against all vectors

        Args:
            query: query string to be used for vector search
            top_k: number of results to return
            return_similarities: whether to return similarity values along with documents

        Returns:
           A list of documents 
        """
        if self._collection is None:
            raise Exception("No existing collection. Create one first.")

        query_vector = self._get_embedding(query)
        vector_store = vstack(self._collection.embeddings)

        similarities = self.sim_func(query_vector, vector_store)
        indices = argsort(similarities)[-top_k:][::-1]
        documents = self._collection.documents

        if return_similarities:
            return [(documents[idx], similarities[idx]) for idx in indices]

        return [documents[idx] for idx in indices]

    def _get_embedding(self, text: str):
        return self.embed_model.encode(text)

    def _save_to_disk(self, fname: str):
        with open(os.path.join(self.persist_dir, fname), "wb") as f:
            dump(self._collection, f, HIGHEST_PROTOCOL)
